import traceback
import time
import asyncio
from miniopy_async import Minio
from miniopy_async.lifecycleconfig import LifecycleConfig, Rule, Expiration
from miniopy_async.commonconfig import ENABLED, DISABLED, Filter
from telnetlib import Telnet
from threading import Thread, Lock

from concurrent.futures.thread import ThreadPoolExecutor

class MinioClient(Minio):
    def __init__(self, endpoint, access_key=None, secret_key=None, session_token=None, secure=True, region=None, credentials=None):
        super().__init__(endpoint, access_key, secret_key, session_token, secure, region, credentials)
        self.location = region

    async def create_bucket(self, bucket_name, object_lock=False):
        return await super().make_bucket(bucket_name, self.location, object_lock)

class MinioAgentAsync:
    def __init__(self, endpoints, weights, check_awailable_ts=1, timeout=10):
        self.client_pool = []

        for endpoint in endpoints:
            self.client_pool.append(MinioClient(**endpoint))
        self.endpoints = endpoints
        self.availability = [True] * len(endpoints)
        self.weights = weights
        self.check_availability_ts = check_awailable_ts
        self.timeout = timeout

        self.availability_locker = Lock()

        self.availability_check_thread = Thread(target=self.check_availability, daemon=True)
        self.availability_check_thread.start()

    def check(self, endpoint):
        try:
            host, port = endpoint['endpoint'].split(':')
            with Telnet(host, int(port), timeout=3) as tn:
                return True
        except Exception as e:
            print(e)
            return False

    def check_availability(self):
        while True:
            print('checking endpoints availability...')
            self.availability_locker.acquire(blocking=True, timeout=30)
            try:
                with ThreadPoolExecutor(max_workers=2) as pool:
                    results = pool.map(self.check, self.endpoints)
                    for i, r in enumerate(results):
                        print('endpoint {} -> {}'.format(i, r))
                        self.availability[i] = r

                        if not r:
                            self.client_pool[i] = None
                        else:
                            self.client_pool[i] = MinioClient(**self.endpoints[i])
            except Exception as e:
                print(e)
            finally:
                self.availability_locker.release()
                time.sleep(self.check_availability_ts)

    def get_client(self) -> MinioClient:
        self.availability_locker.acquire(blocking=True, timeout=30)

        fused_weights = [(i, w) for i, (a, w) in enumerate(zip(self.availability, self.weights)) if a]
        fused_weights = sorted(fused_weights, key=lambda x:x[1], reverse=True)        
        self.availability_locker.release()
        print('using endpoint -> {}'.format(self.endpoints[fused_weights[0][0]]['endpoint']))
        return self.client_pool[fused_weights[0][0]]

    async def create_bucket_if_not_exists(self, bucket_name, lifecycle=None):
        try:
            if not await self.get_client().bucket_exists(bucket_name):
                await self.get_client().create_bucket(bucket_name)

                if lifecycle is not None:
                    await self.get_client().set_bucket_lifecycle(bucket_name, LifecycleConfig([
                        Rule(ENABLED, expiration=Expiration(days=lifecycle), rule_filter=Filter(prefix=''))
                    ]))

            return True
        except Exception as e:
            if 'Bucket name contains invalid characters' in str(e):
                return False
            traceback.print_exc()
            return True

    async def _upload_file(self, path, bucket, bucket_filename, tries=0, max_tries=5):
        try:
            if await self.create_bucket_if_not_exists(bucket):
                await self.get_client().fput_object(bucket, bucket_filename, path)
                return True, 'minio://{}/{}'.format(bucket, bucket_filename)
            else:
                print('創建bucket失敗: {}'.format(bucket))
                return False, '創建bucket失敗'
        except Exception as e:
            if tries > max_tries:
                return False, e
            else:
                return await self._upload_file(path, bucket, bucket_filename, tries+1)
            
    async def upload_file(self, path, bucket, bucket_filename, max_tries=5):
        # future = self.upload_pool.submit(self._upload_file, path, bucket, bucket_filename)
        # return future.result(timeout=10)
        try:
            return await asyncio.wait_for(self._upload_file(path, bucket, bucket_filename, max_tries=max_tries), timeout=self.timeout)
        except Exception as e:
            return False, str(e)

    async def _upload_file_by_buffer(self, buffer, bucket, bucket_filename, tries=0):
        try:
            if await self.create_bucket_if_not_exists(bucket):
                await self.get_client().put_object(
                    bucket, bucket_filename, buffer, len(buffer.getvalue()))
                return True, 'minio://{}/{}'.format(bucket, bucket_filename)
            else:
                return False, '創建bucket失敗'
        except Exception as e:
            if tries > 3:
                return False, e
            else:
                return await self._upload_file_by_buffer(buffer, bucket, bucket_filename, tries+1)
            
    async def upload_file_by_buffer(self, buffer, bucket, bucket_filename):
        # future = self.upload_pool.submit(self._upload_file_by_buffer, buffer, bucket, bucket_filename)
        # return future.result(timeout=10)
        try:
            return await asyncio.wait_for(self._upload_file_by_buffer(buffer, bucket, bucket_filename), timeout=self.timeout)
        except Exception as e:
            return False, str(e)

    async def _download_file(self, path, bucket, bucket_filename, tries=0, max_tries=5):
        try:
            await self.get_client().fget_object(bucket, bucket_filename, path)
            return True, path
        except Exception as e:
            if tries > max_tries:
                return False, e
            else:
                return await self._download_file(path, bucket, bucket_filename, tries+1)
            
    async def download_file(self, path, bucket, bucket_filename, max_tries=5):
        # future = self.download_pool.submit(self._download_file, path, bucket, bucket_filename)
        # return future.result(timeout=10)
        try:
            return await asyncio.wait_for(self._download_file(path, bucket, bucket_filename, max_tries=max_tries), timeout=self.timeout)
        except Exception as e:
            return False, str(e)

    async def delete(self, bucket, bucket_name, tries):
        try:
            await self.get_client().remove_object(bucket, bucket_name)
            return True, None
        except Exception as e:
            if tries > 5:
                return False, e
            else:
                return await self.delete(bucket, bucket_name, tries + 1)

    async def exists(self, bucket, bucket_name):
        meta = await self.get_client().stat_object(bucket, bucket_name)

        return meta is not None



import orjson
import asyncio
import nats
from nats.errors import ConnectionClosedError, TimeoutError, NoServersError
from hamunafs.utils.singleton_wrapper import Singleton

class NATSClient(Singleton):
    def __init__(self, host, port):
        self.host = host
        self.port = port

    async def connect(self):
        self.client = await nats.connect('nats://{}:{}'.format(self.host, self.port))

    async def publish(self, topic, params):
        if isinstance(params, str):
            _params = params.encode()
        else:
            _params = orjson.dumps(params)
        await self.client.publish(topic, _params)



from setuptools import setup, find_packages

setup(
    name='hmfsv2',
    version='1.0.7',
    packages=find_packages(),
    license="MIT",
    description='Distributed filesystem for Hamuna AI',
    long_description='Distributed filesystem for Hamuna AI with multiple third party middleware transfer points',
    long_description_content_type="text/plain",
    author='O.Push',
    author_email='opush.developer@outlook.com',
    url='https://www.hamuna.club',
    package_dir={'': '.'},
    install_requires=['minio', 'redis==4.3.4', 'paho-mqtt', 'gnsq', 'ansq', 'qiniu', 'boto', 'diskcache',
                      'httpx', 'aiofile', 'aiohttp', 'aiohttp_retry', 'aiobotocore', 'nos-python3-sdk', 'miniopy_async', 'aiodecorators', 'wget']
)




import traceback
from gnsq import Producer
from ansq import open_connection, ConnectionOptions
import asyncio
import orjson
import time
from aiodecorators import Semaphore

from hamunafs.utils.singleton_wrapper import Singleton


class MQManager(Singleton):
    def __init__(self, host, port, async_mq=False, init=True):
        if not self.need_init():
            return
        self.host = host
        self.port = port
        self.async_mq = async_mq
        self.inited = False
        if init:
            self._init()
        self._inited = True

        self.connection_locker = asyncio.Lock()
    
    def _init(self):
        self.addr = '{}:{}'.format(self.host, self.port)
        self.producer = Producer(nsqd_tcp_addresses=[self.addr])
        self.producer.start()

    async def _init_async(self):
        if self.async_mq:
            self.mq = await open_connection(self.host, self.port, connection_options=ConnectionOptions(debug=True))

    def close(self):
        if not self.async_mq:
            self.close_producer()

    async def close_async(self):
        try:
            if self.async_mq:
                try:
                    await self.mq.close()
                    del self.mq
                except:
                    pass
        except:
            traceback.print_exc()

    async def get_mq_conn(self):
        # async with self.connection_locker:
            # if not hasattr(self, 'mq') or self.mq.status.is_closed:
            #     await self._init_async()
            #     await asyncio.sleep(0.1)
            # else:
            #     if not self.mq.status.is_connected:
            #         is_timeout = False
            #         t = time.time()
            #         while self.mq.status.is_reconnecting:
            #             if time.time() - t > 5:
            #                 is_timeout = True
            #                 break
            #             await asyncio.sleep(0.1)
                    
            #         if self.mq.status.is_connected:
            #             return self.mq
                    
            #         if is_timeout:
            #             await self.close_async()
            #             return None
        async with self.connection_locker:
            if not hasattr(self, 'mq'):
                await self._init_async()
                await asyncio.sleep(0.1)
        return self.mq
            
    def _encode_message(self, message):
        if isinstance(message, str):
            message = message.encode('utf-8')
        else:
            message = orjson.dumps(message)
        return message

    def close_producer(self):
        try:
            self.producer.close()
        except:
            pass 
        
    def restart_producer(self):
        try:
            self.close_producer()
            del self.producer
        except:
            pass
        finally:
            self._init()
            
    def publish(self, topic, message, multi=False):
        try:
            if not self.producer.is_running:
                self.producer.close()
                time.sleep(0.5)
                self._init()
                
            if multi and isinstance(message, list):
                message = list(map(self._encode_message, message))
            else:
                message = self._encode_message(message)
            
            if multi:
                pub_func = self.producer.multipublish
            else:
                pub_func = self.producer.publish
            
            ret = pub_func(topic, message).decode()
        except Exception as e:
            traceback.print_exc()
            ret = 'ERR'
        max_t = 5
        t = 0
        while ret != 'OK' and t < max_t: 
            try:
                ret = pub_func(topic, message).decode()
            except:
                ret = 'ERR'
            if ret == 'ERR':
                time.sleep(1)
            t += 1
        ret = ret == 'OK'
        if ret:
            print('nsq publish success!')
        return ret

    async def __mq_pub(self, topic, message, mq=None, tries=0):
        if tries > 5:
            return False
        if mq is None:
            mq = await self.get_mq_conn()
        try:
            if mq is None:
                return await self.__mq_pub(topic, message, None, tries+1)
                
            resp = await mq.pub(topic, message)
            return resp.is_ok
        except:
            traceback.print_exc()
            await self.close_async()
            return await self.__mq_pub(topic, message, None, tries+1)

    @Semaphore(8)
    async def async_publish(self, topic, message, mq=None):
        try:
            if isinstance(message, str):
                message = message.encode('utf-8')
            elif isinstance(message, bytes):
                pass
            else:
                message = orjson.dumps(message)

            ret = await self.__mq_pub(topic, message)
            if ret:
                print('published')
            else:
                print('publish failed')
            return ret
        except Exception as e:
            print(traceback.print_exc())
            return False

if __name__ == "__main__":
    manager = MQManager('kafka.ai.hamuna.club', 34150, async_mq=True)
    
    loop = asyncio.get_event_loop()
    loop.run_until_complete(manager._init_async())
    
    for i in range(100):
        ret = loop.run_until_complete(manager.async_publish('test', 'test'))
        # ret = manager.publish('test', 'test'.encode('utf-8'))
        print(ret)